/*
   LinuxHandlerDomain.c : Linux handler domain.
   Copyright (C) 2001 Karim Yaghmour.
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   History : 
     K.Y., 10/02/2001, Initial typing.
*/

/* Necessary defines */
#define __KERNEL__
#define MODULE

/* Module and initialization stuff */
#include <linux/module.h>
#include <linux/init.h>

/* Necessary includes */
#include <asm/processor.h>
#include <asm/pgtable.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/types.h>

/* Local definitions */
#include "LinuxHandlerDomain.h"

/* Gdt/idt limit and base */
typedef struct _dt_limit_base
{
  u16    limit;
  u32    base __attribute__((packed));
} dt_limit_base;

/* Intel x86 descriptor */
typedef struct _descriptor
{
  u32    part1;
  u32    part2 __attribute__((packed));
} descriptor;

/* Local variables */
static dt_limit_base      orig_gdt = {0,0}; /* Original GDT limit and base */
static dt_limit_base      orig_idt = {0,0}; /* Original IDT limit and base */

/******************************************************************
 * Function :
 *    init_module()
 * Description :
 *    Standard loadable module initialization function.
 * Parameters :
 *    NONE
 * Return values :
 *    0, if everything went OK.
 * History :
 *    K.Y., 10/02/2001, Initial typing.
 * Note :
 ******************************************************************/
int init_module(void)
{
  int         i;             /* Generic index */
  u16         cur_tr;        /* Current task register */
  u32         cur_ldt_start; /* Current LDT start */
  u64*        gdt = 0;       /* Pointer to GDT entry */
  u64*        idt = 0;       /* Pointer to IDT entry */
  descriptor* desc_ptr;      /* Descriptor pointer */

  /* LHD now firing up */
  printk("LHD: Loading Linux handler domain ... \n");

  /* Take control of the hardware */
  __asm__ __volatile__("cli");

  /* Save the Global Descriptor Table */
  __asm__ __volatile__("sgdt %0": "=m" (orig_gdt));

  /* Save the Interrupt Descriptor Table */
  __asm__ __volatile__("sidt %0": "=m" (orig_idt));

  /* Read the Local Descriptor Table */
  __asm__ __volatile__("sldt %0": "=m" (cur_ldt_start));

  /* Read the current task register */
  __asm__ __volatile__("str %0": "=m" (cur_tr));

  /* Get pointer to first GDT and IDT entries */
  gdt = (u64*) orig_gdt.base;
  idt = (u64*) orig_idt.base;

  /* Don't crash the machine for now: Set IOPL = 1 */
  __asm__ __volatile__("pushfl; orl $0x00001000,(%esp); popfl");

  /* Take hardware control away from Linux by pushing Linux's code into ring 1 */
  desc_ptr = (descriptor*) gdt + 2;
  desc_ptr->part2 |= 0x00004000;
  desc_ptr = (descriptor*) gdt + 3;
  desc_ptr->part2 |= 0x00004000;

  /* Set the interrupt descriptor DPL's to 01 */
  for(i = 0, desc_ptr = (descriptor*) idt;
      i < 256;
      i++, desc_ptr = (descriptor*) (idt + i))
    desc_ptr->part2 |=00004000;    

#define __STR(x) #x
#define STR(x) __STR(x)

  /* Reload all segment registers after changing GDT */
  __asm__ __volatile__("ljmp 1      \n"
		       "1: movl $" STR(__KERNEL_DS)",%eax;     \n"
		       "movl %eax,%ds;                \n"
		       "movl %eax,%es;                \n"  
		       "movl %eax,%fs;                \n"
		       "movl %eax,%gs;                \n"
		       "movl $" STR(__KERNEL_DS)", %eax;    \n"
		       "movl %eax,%ss;                \n"
		       "cld		              \n");

  /* Reload TSS */
  __asm__ __volatile__("ltr %0": "=m" (cur_tr));

  /* Reload LDT */
  __asm__ __volatile__("lldt %0": "=m" (cur_ldt_start));

  /* Invalidate cache */
  __asm__ __volatile__("invd");

  /* Invalidate TLB entries */
  do{
    unsigned int tmpreg;
    __asm__ __volatile__("movl %%cr4, %%eax; \n"
			 "xorl $0x00000080,%%eax \n"
			 "movl %%eax,%%cr4 \n"
			 "movl %%cr3, %0; \n"
			 "movl %0, %%cr3; \n"
			 "movl %%cr4, %%eax; \n"
			 "xorl $0x00000080,%%eax; \n"
			 "movl %%eax,%%cr4; \n"	 
			 : "=r" (tmpreg)
			 :: "memory");
    }
  while(0);

  /* Let go of the hardware */
  __asm__ __volatile__("sti");

  /* ??? Maybe I shouldn't be calling any of these printk's as they aren't in ring-zero anymore  ??? */

  /* Print out some info regarding the system */
  printk("LHD: Original GDT at virt addr: 0x%08X \n", orig_gdt.base);
  printk("LHD: Original GDT at phys addr: 0x%08X \n", virt_to_phys((void*) orig_gdt.base));
  printk("LHD: Orignial GDT limit:        %d \n", orig_gdt.limit);

  printk("\n");

  printk("LHD: Original IDT at virt addr: 0x%08X \n", orig_idt.base);
  printk("LHD: Original IDT at phys addr: 0x%08X \n", virt_to_phys((void*) orig_idt.base));
  printk("LHD: Orignial IDT limit:        %d \n", orig_idt.limit);

  printk("\n");

  printk("LHD: Current  LDT at phys addr: 0x%08X \n", cur_ldt_start);
  printk("LHD: Current  TR  contains:     0x%04X \n", cur_tr);

  printk("\n");

  /* Print out first GDT entries */
  for(i = 0, desc_ptr = (descriptor*) gdt;
      i < 12;
      i++, desc_ptr = (descriptor*) (gdt + i))
    printk("GDT entry # %2d  => 0x%08X%08X \n", i, desc_ptr->part2, desc_ptr->part1);

  printk("\n");

  /* Print out first IDT entries */
  for(i = 0, desc_ptr = (descriptor*) idt;
      i < 12;
      i++, desc_ptr = (descriptor*) (idt + i))
    printk("IDT entry # %2d  => 0x%08X%08X \n", i, desc_ptr->part2, desc_ptr->part1);

  /* Print out IRQ IDT entries */
  for(i = 32, desc_ptr = (descriptor*) idt;
      i < 48;
      i++, desc_ptr = (descriptor*) (idt + i))
    printk("IDT entry # %2d  => 0x%08X%08X \n", i, desc_ptr->part2, desc_ptr->part1);  

  /* We're done initializing the LHD */
  printk("\nLHD: LHD active! \n");

  /* DEBUG: Just in case return stack causes problems */
#if 0
  while(1);
#endif

  /* Tell the caller that everything went OK */
  return 0;
}

/******************************************************************
 * Function :
 *    cleanup_module()
 * Description :
 *    Standard loadable module cleanup function.
 * Parameters :
 *    NONE
 * Return values :
 *    NONE.
 * History :
 *    K.Y., 10/02/2001, Initial typing.
 * Note :
 ******************************************************************/
void cleanup_module(void)
{
  /* We're out of here */
  printk("LHD: Unloading linux handler domain ... \n");

  /* Take control of the hardware */
  __asm__ __volatile__ ("cli");

  /* Let go of the hardware */
  __asm__ __volatile__ ("sti");

  /* We're out of here */
  printk("LHD: LHD unloaded. \n");
}

